/* SOD32 the Stack Oriented Design computer.
   Copyright 1994, L.C. Benschop, Eindhoven, The Netherlands.
   The program is released under the GNU General Public License version 2.
   There is NO WARRANTY.
*/

#ifdef WIN32
	#include <winsock.h>
#else
	#include <termios.h>
	#include <sys/socket.h>
	#include <sys/ioctl.h>
	#include <signal.h>
	#define closesocket close
	#define ioctlsocket ioctl
#endif

#include <stdio.h>
#include <signal.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <pthread.h>

#include <sys/types.h>
#include <stdlib.h>
#include <dirent.h>

extern char inet_addr_buf[];

#define UNS8  unsigned char
#define UNS16 unsigned short
#define INT32 int
#define UNS32 unsigned int
#define UNS64 unsigned long long

#define MEMSIZE (1<<18) /* must be a power of two */
#define MEMMASK MEMSIZE-1 /* mask for addresses to force them into range */

#define MIN(X,Y) ((X) < (Y) ? (X) : (Y))
#define MAX(X,Y) ((X) > (Y) ? (X) : (Y))

#define  TIB_SAZE 256

typedef struct {
	UNS32 oSTART;
	UNS32 oSP;
	UNS32 oTO_IN;
	UNS32 oNTIB;
	UNS8  oTTIB[TIB_SAZE];
	UNS32 oATIB;
	UNS32 oSTATE;
	UNS32 oHANDLER;
	UNS32 ooverflow;
	UNS32 oDPOINT;
	UNS32 oLAST;
	UNS32 oVFORTH;
	UNS32 oCONTEXT[17];
	UNS32 oCURRENT;
	UNS32 oOUTPUT;
	UNS32 oINPUT;
/*
#define LST	(*(UNS32*)(mem+4*2))
#define VFORTH	(*(UNS32*)(mem+4*3))
	*/
}  f_boot_t ;

f_boot_t * f_boot;

#define farrd(reg)  ((int)&(f_boot->reg) - (int)(f_boot))

#define SP		farrd(oSP)
#define TO_IN		farrd(oTO_IN)
#define NTIB	 	farrd(oNTIB)
#define TTIB	 	farrd(oTTIB)
#define ATIB	 	farrd(oATIB)
#define STATE	 	farrd(oSTATE)
#define HANDLER 	farrd(oHANDLER)
#define overflow	farrd(ooverflow)
#define DPOINT		farrd(oDPOINT)
#define LAST		farrd(oLAST)
#define VFORTH		farrd(oVFORTH)
#define CONTEXT		farrd(oCONTEXT)
#define CURRENT		farrd(oCURRENT)
#define OUTPUT		farrd(oOUTPUT)
#define INPUT		farrd(oINPUT)

#define vocs_beg 	sizeof(*f_boot)


#ifdef WIN32

#define WCELL(reg) (*(UNS16*)((UNS32)mem+(reg&MEMMASK)))
#define CELL(reg)  (*(UNS32*)((UNS32)mem+(reg&MEMMASK)))
#define DCELL(reg) (*(UNS64*)((UNS32)mem+(reg&MEMMASK)))

#define FSTART	(*(UNS32*)((UNS32)mem))
char * HPATH="HOMEPATH";

#else

#define WCELL(reg) (*(UNS16*)&mem[reg&MEMMASK])
#define CELL(reg)  (*(UNS32*)&mem[reg&MEMMASK])
#define DCELL(reg) (*(UNS64*)&mem[reg&MEMMASK])

#define FSTART	(*(UNS32*)&mem[0])
char * HPATH="HOME";

#endif

#define HERE 	CELL(DPOINT)
//   HERE ((*(f_boot_t*)&mem).oDPOINT)

#define last 	CELL(LAST)
//#define vforth 	CELL(VFORTH)
#define context 	CELL(CELL(CONTEXT))
#define current 	CELL(CELL(CURRENT))
#define TIB 	CELL(ATIB)



#define TIB 	CELL(ATIB)


/* Key for the thread-specific buffer */
 static pthread_key_t buffer_key;

 /* Once-only initialisation of the key */
 static pthread_once_t buffer_key_once = PTHREAD_ONCE_INIT;


//UNS8 mem[MEMSIZE+3+4];

#if 0
#if 1
#define mem ((UNS8 *) pthread_getspecific(buffer_key))
#define sp (*(UNS32*)&mem[SP])
#else
volatile register UNS32 sp;
volatile register UNS8 * mem;
#endif
#endif

#define MEMDEF register UNS8 * mem = ((UNS8 *) pthread_getspecific(buffer_key));
#define sp (*(UNS32*)&mem[SP])


UNS32 save_ip,interrupt;
//UNS32 netid;

#define BYTE(reg) mem[reg&MEMMASK]

#define SWAP ;
#define CLIP addr&=MEMMASK;len&=MEMMASK;if(addr+len>MEMSIZE)len=MEMSIZE-addr;
#define QFERR if(!fp)  {ior=-201;CELL(sp)=ior;break;}
#define FILEID(n) if((n)>=20){ior=-201;CELL(sp)=ior;break;} fp=fileids[n];


//#define PUSH(n)	(sp-=4,CELL(sp)=n)
//#define POP()	(CELL(sp),sp+=4)

char filename[256];

static struct termios oldterm;
static struct termios newterm;
static int tflags;
static int realterm;
int keybuf,raw;
static int childpid;

#ifndef O_NDELAY
#define O_NDELAY O_NONBLOCK
#endif

void inthandler(int s)
{
 signal(SIGINT,inthandler);
 interrupt=12;
}

void alarmhandler(int s)
{
 interrupt=16;
}

#ifndef WIN32

char key_query_char = 0;


int getch(void)
{
	  char ch;
	  int n;
	  struct termios t1, t2;

	      tcgetattr(0, &t1);
	      t2 = t1;
	      t2.c_lflag &= ~ICANON;
	      t2.c_lflag &= ~ECHO;
	      t2.c_lflag |= ISIG;
	      t2.c_cc[VMIN] = 1;
	      t2.c_cc[VTIME] = 0;
	      tcsetattr(0, TCSANOW, &t2);

		n = read(0, &ch, 1);

	      tcsetattr(0, TCSANOW, &t1);

	  return ch;
}

int kbhit() {
	  struct termios t1, t2;
	    int bytesWaiting;

    tcgetattr(0, &t1);

      	t2 = t1;
        t2.c_lflag &= ~ICANON;

    tcsetattr(0, TCSANOW, &t2);

    	ioctl(0, FIONREAD, &bytesWaiting);

    tcsetattr(0, TCSANOW, &t1);

    return bytesWaiting;
}


void putch(int c)
{
	char k=c;
	while(write(1,&k,1)<0);
}


SIGINTTypt()
{
	printf("SIGALRM\n");
}


SIGALRMTypt()
{
	printf("SIGALRM\n");
}

SIGSIGPIPETypt()
{
	printf("disconnect\n");
}

SIGSIGTERMTypt()
{
	printf("SIGSIGTERM\n");
}
SIGSIGSEGVTypt()
{
	printf("SIGSIGSEGV\n");
}


#endif

int make_name(char *addr,UNS32 len)
{
 int i;
 for(i=0;i<len && i<256;i++) {
    if(addr[i]=='.' && addr[i+1]=='.') return 0;
    filename[i]=addr[i];
   /* Prohibit filenames in other directories */
 }
 filename[i]='\0';
 return 1;
}


load_image(char *name)
{
 UNS32 len;
 FILE *infile;
 MEMDEF
 if((infile=fopen(name,"rb"))==NULL) {
  fprintf(stderr,"Cannot open image file.\n");
  exit(2);
 }
 len=fread(mem,1,MEMSIZE,infile);
}


/* Multiply 32-bit unsigned numbers *a and *b.
   High half of 64-bit result in *a, low half in *b
*/
static void umul(UNS32 *a,UNS32 *b)
{
 UNS32 ah,al,bh,bl,ph,pm1,pm2,pl;
 ah=*a>>16;al=*a&0xffff;
 bh=*b>>16;bl=*b&0xffff;
 pl=al*bl;
 if((ah|bh)==0) {
  ph=0;
 } else {
  pm1=al*bh;
  pm2=ah*bl;
  ph=ah*bh;
  pl=pl+(pm1<<16);
  ph+=(pl<(pm1<<16));
  pl=pl+(pm2<<16);
  ph+=(pl<(pm2<<16));
  ph=ph+(pm1>>16)+(pm2>>16);
 }
 *a=ph;*b=pl;
}

/* Divide 64-bit unsigned number (high half *b, low half *c) by
   32-bit unsigend number in *a. Quotient in *b, remainder in *c.
*/
static void udiv(UNS32 *a,UNS32 *b,UNS32 *c)
{
 UNS32 d,qh,ql;
 int i,cy;
 qh=*b;ql=*c;d=*a;
 if(qh==0) {
  *b=ql/d;
  *c=ql%d;
 } else {
  for(i=0;i<32;i++) {
   cy=qh&0x80000000;
   qh<<=1;
   if(ql&0x80000000)qh++;
   ql<<=1;
   if(qh>=d||cy) {
    qh-=d;
    ql++;
    cy=0;
   }
   *c=qh;
   *b=ql;
  }
 }
}

enum {
sw_nop,
sw_swap,
sw_rot,
sw_0equ,
sw_0Nequ,
sw_0LT,
sw_0GT,
sw_equ,
sw_nequ,
sw_negate,
sw_invert,
sw_ummul,
sw_Cload,
sw_Wload,
sw_load,
sw_1add,
sw_2add,
sw_1sub,
sw_2sub,
sw_2mul,
sw_2div,
sw_add,
sw_sub,
sw_mul,
sw_div,
sw_mod,
sw_celldiv,
sw_cells,
sw_and,
sw_or,
sw_xor,
sw_uLT,
sw_LT,
sw_uGT,
sw_GT,
sw_min,
sw_max,
sw_lshift,
sw_rshift,
sw_umSmod,
sw_addcy,
sw_Dadd,
sw_scan1,
sw_drop,
sw_2drop,
sw_on,
sw_0ST,
sw_ST,
sw_addST,
sw_cST,
sw_wST,
sw_2ST,
sw_2load,
sw_2swap,
sw_cSTa,
sw_qdup,
sw_dup,
sw_2dup,
sw_over,
sw_rload,
sw_rGT,
sw_GTr,
sw_ip,
sw_lit,
sw_litcomp,
sw_comp,
sw_Wcomp,
sw_Ccomp,
sw_allot,
sw_execute,
sw_compile,
sw_immediate,
sw_FromName,
sw_ret,
sw_femd,
sw_spload,
sw_spST,
sw_rpload,
sw_rpST,
sw_to_catch,
sw_from_catch,
sw_throw,
sw_setalarm,
sw_bye,
sw_branch,
sw_0branch,
sw_BEGIN,
sw_AGAIN,
sw_UNTIL,
sw_AHEAD,
sw_IF,
sw_WHILE,
sw_THEN,
sw_ELSE,
sw_REPEAT,
sw_key,
sw_emit,
sw_rdch,
sw_keyq,
sw_type,
sw_hdot,
sw_decdot,
sw_dir,
sw_count,
sw_str,
sw_StrComp,
sw_NextWord,
sw_Parse,
sw_SQ,
sw_PQ,
sw_Source,
sw_SourceST,
sw_SFind,
sw_SearchWordList,
sw_SHeader,
sw_openfile,
sw_closefile,
sw_writeline,
sw_readline,
sw_writefile,
sw_readfile,
sw_deletefile,
sw_repositionfile,
sw_fileposition,
sw_system,
sw_filesize,
sw_restore,
sw_telid,
sw_ipadr,
sw_connect,
sw_tellen,
sw_telp,
sw_ms,
sw_hi
};


void COM(UNS32 dd)
{ MEMDEF
 CELL(HERE)=dd;
HERE+=4;
}

void WCOM(UNS16 dd)
{ MEMDEF
	WCELL(HERE)=dd;
	HERE+=2;
}

void CCOM(UNS8 dd)
{ MEMDEF
 BYTE(HERE)=dd;
  HERE+=1;
}

void StrComp(UNS8 * word,UNS32 size)
{ UNS32 addr1,addr0;
  MEMDEF
addr0=HERE;
addr1=addr0+1;
while(size--)
	   BYTE(addr1++)=*word++;
BYTE(addr0)=addr1-addr0-1;
HERE=addr1;
if(HERE&1)CCOM(0);
}

void SHEADER(UNS8 * word,UNS32 size)
{       MEMDEF
	if(current)  WCOM(current-HERE);
	else		WCOM(0);
   last=HERE;
   StrComp(word,size);
}

void HEADER(UNS8 * word)
{  MEMDEF
	size_t i;
    for (i = 0; word[i] != 0; i ++);
	SHEADER(word,i);
}

void IMMED()
{       MEMDEF
	BYTE(last)=BYTE(last)|0x80;
}

void PRIM(UNS8 * word,UNS16 cod)
{	MEMDEF
	HEADER(word);
	current=last;
	WCOM(cod<<1|0x8001);
}

void CONSTANT(char * word,UNS32 dd)
{	MEMDEF
	HEADER(word);
	current=last;
	WCOM(sw_lit<<1|0x4001);
	COM(dd);
}

void CREATE(char * word)
{	MEMDEF
	HEADER(word);
	current=last;
	WCOM(sw_ip<<1|0x4001);
}

void VARIABLE(char * word)
{	MEMDEF
	CREATE(word);
	COM(0);
}

UNS32 SEARCH(UNS32 wid,  char * word)
{ UNS32 addr=wid;
  MEMDEF
  for(;;)
  { UNS16 link;
	if( !memcmp(word,&BYTE(addr+1),BYTE(addr)&0x7F) )
	return addr;
	addr-=2;
	link=WCELL(addr);
	if(!link) return 0;
	addr+=(link^0x8000)-0x8000;
  }
}

UNS32 FromName(UNS32 addr)
{ MEMDEF
 addr+=(BYTE(addr)&0x7f)+1;
  if(addr&1)addr++;
  return addr;
}

void Compile(UNS32 addr)
{	UNS32 cod;
	MEMDEF
	  cod=WCELL(addr);
	  if((cod&0x8001)==0x8001)
		  {WCOM(cod&0x3FFF); return;}
	  WCOM(addr-(HERE+2));

}
void CC(char * word)
{       MEMDEF
	UNS32 addr=current;
	UNS8 size;

	for (size = 0; word[size] != 0; size ++);

	do	{
		addr=SEARCH(addr,word);
	   	if((BYTE(addr)&0x3f)==size ) break;
		addr-=2;
		addr+=(WCELL(addr)^0x8000)-0x8000;
	}while(addr);

	  if(!addr)
	    for(;;);
	  Compile(FromName(addr));
}

void STR(char * str)
{	MEMDEF
	char * str0=str;
	UNS32 addr0;
	WCOM(sw_str<<1|1);
	addr0=HERE;
	CCOM(0);
	while(*str)  CCOM(*str++);
        BYTE(addr0)=(UNS8)(str-str0);
	if(HERE&1)CCOM(0);
}

void TP(char * str)
{ MEMDEF
 STR(str);
WCOM(sw_type<<1|1);
}

void F_END()
{	MEMDEF
	WCOM(0x8001);
	current=last;
}

void BEGIN()
{	MEMDEF
	sp-=4;CELL(sp)=HERE;
// printf("BEGIN=%x\n",CELL(sp));
}

void AGAIN()
{	MEMDEF
	WCOM(sw_branch<<1|1);
//printf("AGAIN=%x\n",CELL(sp));
	WCOM(CELL(sp)-HERE);
	sp+=4;
}

void UNTIL()
{	MEMDEF
	WCOM(sw_0branch<<1|1);
	WCOM(CELL(sp)-HERE);
	sp+=4;
}

void AHEAD()
{       MEMDEF
	sp-=4;CELL(sp)=0;
	AGAIN();
	sp-=4;CELL(sp)=HERE-2;
}

void IF()
{       MEMDEF
	sp-=4;CELL(sp)=0;
	UNTIL();
	sp-=4;CELL(sp)=HERE-2;
}

void WHILE()
{ MEMDEF
  UNS32 t;
	IF();
	t=CELL(sp);
	CELL(sp)=CELL(sp+4);
	CELL(sp+4)=t;
}

void THEN()
{       MEMDEF
	WCELL(CELL(sp))=HERE-CELL(sp);
	sp+=4;
}

void ELSE()
{ MEMDEF
  UNS32 t;
	AHEAD();
	t=CELL(sp);
	CELL(sp)=CELL(sp+4);
	CELL(sp+4)=t;
	THEN();
}

void REPEAT()
{	MEMDEF
	AGAIN();
	THEN();
}

void LIT(UNS32 dd)
{	MEMDEF
	WCOM(sw_lit<<1|1);
	COM(dd);
}

void NextWord()
{ MEMDEF
  UNS32 addr,Waddr,Eaddr;
addr=CELL(TO_IN) + TIB ;
Eaddr=TIB+CELL(NTIB);
sp-=8;
	while ( (BYTE(addr) <= ' ') & (addr<Eaddr ) )addr++;
	CELL(sp+4)=Waddr=addr;
	while ( (BYTE(addr++)  > ' ') & (addr<Eaddr )	);
	CELL(sp)=addr-Waddr-1;
	CELL(TO_IN)=addr - TIB;
}

void Parse()
{ MEMDEF
  UNS32 addr,Waddr,Eaddr;
   UNS8 ch;
addr=CELL(TO_IN) + TIB ;
Eaddr=TIB+CELL(NTIB);
ch=BYTE(sp);
sp-=4;
	CELL(sp+4)=Waddr=addr;
	while ( (BYTE(addr++)  != ch) & (addr<Eaddr )	);
	CELL(sp)=addr-Waddr-1;
	CELL(TO_IN)=addr - TIB;
}

void SQ()
{	MEMDEF
	sp-=4; CELL(sp)=(UNS32)'"';Parse();
	if(CELL(STATE))
	{		UNS32 addr,size;
	WCOM(sw_str<<1|1);
	addr=CELL(sp+4);
 	   size=CELL(sp);
 	   size=MIN(size, MEMSIZE-addr);
 	   StrComp(&BYTE(addr),size);
  sp+=8;
	}
}
void Source()
{	MEMDEF
	sp-=8;
	CELL(sp+4)=TIB;
	CELL(sp)=CELL(NTIB);
}

void SourceST()
{	MEMDEF
	UNS32 len,addr;
	CELL(NTIB)=len=CELL(sp);
	addr=CELL(sp+4);
	while(len--)	BYTE(TIB+len)=BYTE(addr+len);
	sp+=8;
}

void SearchWordList() // ( c-addr u wid --- 0 | xt 1 xt -1 )
{	MEMDEF
  UNS32 addr=CELL(CELL(sp));
   UNS8  size=CELL(sp+4);
   char * word=&BYTE(CELL(sp+8));
   for(;;)
    {  	addr=SEARCH(addr,word);
    	if(!addr)
    	{
    		sp+=8; CELL(sp)=0; return;
    	}
    	if((BYTE(addr)&0x3f)==size )
    	{ sp+=4;
    	  if(BYTE(addr)&0x80)
    	   CELL(sp)=1;
    	  else
       	   CELL(sp)=-1;
    	  CELL(sp+4)=FromName(addr);
    	  return;
    	}
		addr-=2;
		addr+=(WCELL(addr)^0x8000)-0x8000;
    }

}

void SFind()
{	int voc=CONTEXT;
	MEMDEF
	sp-=4; CELL(sp)=0;
	while( CELL(voc) )
	{	sp+=4;
	    sp-=4;CELL(sp)=CELL(sp+8);
	    sp-=4;CELL(sp)=CELL(sp+8);
		sp-=4;CELL(sp)=CELL(voc);
		SearchWordList();
		if(CELL(sp))
		{
			CELL(sp+8)=CELL(sp);
			CELL(sp+12)=CELL(sp+4);
			sp+=8;
			return;
		}	 voc+=4;
	}
}

ssize_t twrite (int __fd, __const void *__buf, size_t __n)
{	MEMDEF
	if(__fd&0x80000000) return(send(__fd^0x80000000,__buf,__n,0));
	else	return(write(__fd,__buf,__n));
}

ssize_t tread (int __fd, __const void *__buf, size_t __n)
{ int nn;
	if(__fd&0x80000000)	return(recv(__fd^0x80000000,__buf,__n,0));
	else	return(read(__fd,__buf,__n));
}

#ifndef eof
int eof(int __fd)
{ int fpt,flg;
    fpt=lseek(__fd, (off_t) 0, SEEK_CUR);
    flg=(fpt==lseek(__fd, (off_t) 0, SEEK_END));
       lseek(__fd, (off_t) fpt, SEEK_SET);
       return flg;
 }
#endif

void virtual_machine(UNS32 netid)
{
	MEMDEF
 register rp,ret,ip,ireg,t;

 UNS32 addr,len,ior=0,ips,CCNN=-1;
 UNS32 zzz=0;
 UNS32 TELID=-1;
 UNS64 dtmp;
 UNS32 backup_mode=0;
 char * fpath="/";
 char	tplast='\n';
 int	calltelp=0;

 int res;
 UNS32 fileids[20]={
  0,0,0,0,0, 0,0,0,0,0,
  0,0,0,0,0, 0,0,0,0,0, };
 INT32 fp;

 fileids[0]=0;
 fileids[1]=1;
 printf("virtual_machine\n");
 if(netid)
 {	 fileids[0]=netid;
	 fileids[1]=netid;
	 fpath="/FPPublic/";
 }
// fileids[0]=0; fileids[1]=1;

// fileids[2]=0;
// fileids[3]=1;

 interrupt=0;
 rp=MEMSIZE;
 ip=FSTART;
 for(;;) {
  if(interrupt) {
   ireg=1;
  doint:
   rp-=4;
  CELL(rp)=ip;
   sp-=4;CELL(sp)=ireg;
   ip=CELL(interrupt);
   interrupt=0;
  }
  ireg=WCELL(ip);
  ips=ip;
  ip+=2;
  //if(netid)  {   twrite(netid, "hello " ,6);  }

  restart: /* Instruction restart after interrupt */
  if(!(ireg&1)) {
            rp-=4;
            CELL(rp)=ip;
            ip+= (ireg^0x8000) - 0x8000;
  } else {
	  if(zzz)  {  printf("z=%x %x %d\n",ip,ireg&0xc000,(ireg&0x3fff)>>1); }
    switch( (ireg&0x3fff)>>1) {
     case sw_swap:	/*swap*/ t=CELL(sp);CELL(sp)=CELL(sp+4);
                      CELL(sp+4)=t;
     case sw_nop:	/*nop*/
    	 break;
     case sw_rot:	/*rot*/  t=CELL(sp+8);CELL(sp+8)=CELL(sp+4);
                      CELL(sp+4)=CELL(sp);CELL(sp)=t;break;
     case sw_0equ:	/*0=*/   CELL(sp)=-!CELL(sp);break;
     case sw_0Nequ:	/*0<>*/   CELL(sp)=-!!CELL(sp);break;
     case sw_0LT:	/*0<*/   CELL(sp)=-((INT32)CELL(sp)<0);break;
     case sw_0GT:	/*0>*/   CELL(sp)=-((INT32)CELL(sp)>0);break;
     case sw_equ:	/*=*/  	 CELL(sp+4)=-(CELL(sp)==CELL(sp+4));sp+=4;break;
     case sw_nequ:	/*=*/  	 CELL(sp+4)=-(CELL(sp)!=CELL(sp+4));sp+=4;break;
     case sw_negate:	/*negate*/ CELL(sp)=-CELL(sp);break;
     case sw_invert:	/*invert*/ CELL(sp)=~CELL(sp);break;
     case sw_ummul:	/* um* */ umul(&CELL(sp),&CELL(sp+4));break;
     case sw_Cload:	/* C@ */  CELL(sp)=BYTE(CELL(sp));break;
     case sw_Wload: /* W@ */  CELL(sp)=WCELL(CELL(sp));break;
     case sw_load:	/* @ */   CELL(sp)=CELL(CELL(sp));break;
     case sw_1add:	/* 1+ */   CELL(sp)++;break;
     case sw_2add:	/* 2+ */   CELL(sp)+=2;break;
     case sw_1sub:	/* 1- */   CELL(sp)--;break;
     case sw_2sub:	/* 2+ */   CELL(sp)-=2;break;
     case sw_2mul:	/* 2* */   CELL(sp)*=2;break;
     case sw_2div:	/* 2/ */   CELL(sp)=(INT32)CELL(sp)/2;break;
     case sw_add:	/* + */   CELL(sp+4)=CELL(sp)+CELL(sp+4);sp+=4;break;
     case sw_sub:	/* - */   CELL(sp+4)-=CELL(sp);sp+=4;break;
     case sw_mul:	/* * */   CELL(sp+4)=(INT32)CELL(sp)*(INT32)CELL(sp+4);sp+=4;break;
     case sw_div:	/* / */   CELL(sp+4)=(INT32)CELL(sp+4)/(INT32)CELL(sp);sp+=4;break;
     case sw_mod:	/* mod */   CELL(sp+4)=(INT32)CELL(sp+4)%(INT32)CELL(sp);sp+=4;break;
     case sw_celldiv:	/* cell/ */ CELL(sp)>>=2;break;
     case sw_cells:	/* cells */ CELL(sp)<<=2;break;
     case sw_and:	/* and */ CELL(sp+4)=CELL(sp)&CELL(sp+4);sp+=4;break;
     case sw_or:	/* or  */ CELL(sp+4)=CELL(sp)|CELL(sp+4);sp+=4;break;
     case sw_xor:	/* xor */ CELL(sp+4)=CELL(sp)^CELL(sp+4);sp+=4;break;
     case sw_uLT:	/* u<  */ CELL(sp+4)=-(CELL(sp+4)<CELL(sp));sp+=4;break;
     case sw_LT:	/* <   */ CELL(sp+4)=-((INT32)CELL(sp+4)<(INT32)CELL(sp));sp+=4;break;
     case sw_uGT:	/* u>  */ CELL(sp+4)=-(CELL(sp+4)>CELL(sp));sp+=4;break;
     case sw_GT:	/* >  */ CELL(sp+4)=-((INT32)CELL(sp+4)>(INT32)CELL(sp));sp+=4;break;
     case sw_min:	/* >  */ CELL(sp+4)=MIN((INT32)CELL(sp+4),(INT32)CELL(sp));sp+=4;break;
     case sw_max:	/* >  */ CELL(sp+4)=MAX((INT32)CELL(sp+4),(INT32)CELL(sp));sp+=4;break;
     case sw_lshift:/*lshift*/CELL(sp+4)<<=CELL(sp);sp+=4;break;
     case sw_rshift:/*rshift*/CELL(sp+4)>>=CELL(sp);sp+=4;break;
     case sw_umSmod:/*um/mod*/if(CELL(sp)<=CELL(sp+4)) { /*overflow */
                         interrupt=overflow;
                         goto doint;
                       }
                       else
                       {
                        udiv(&CELL(sp),&CELL(sp+4),&CELL(sp+8));
			sp+=4;
                       }break;
     case sw_addcy:	/* +cy */ { UNS32 sum; if(CELL(sp)) { /* carry in */
                          sum=CELL(sp+8)+CELL(sp+4)+1;
                          sp+=4;
                          CELL(sp)=(sum<=CELL(sp));
                          CELL(sp+4)=sum;
                         } else {
                          sum=CELL(sp+8)+CELL(sp+4);
                          sp+=4;
                          CELL(sp)=(sum<CELL(sp));
                          CELL(sp+4)=sum;
                         }
                       }break;
     case sw_Dadd:	/* d+ */
     	 	 	 { UNS64 sum= ((UNS64)CELL(sp)<<32) + (UNS64)CELL(sp+4) +
     	 	 		 ((UNS64)CELL(sp+8)<<32) + (UNS64)CELL(sp+12);
                 CELL(sp+8)=sum>>32;
                 CELL(sp+12)=sum; sp+=8;break;
     	 	 	 }
     case sw_scan1:/*scan1*/ if(CELL(sp)){ /* scan from right */
                         t=0;while(!(CELL(sp+4)&1) && t<32) {
                          CELL(sp+4)>>=1;t++;
                         }
                       } else {
                         t=0;while(!(CELL(sp+4)&0x80000000) && t<32) {
                          CELL(sp+4)<<=1;t++;
                         }
                       }sp+=4;CELL(sp)=t;break;
     case sw_drop:/* drop */ sp+=4;break;
     case sw_2drop:/* drop */ sp+=8;break;
     case sw_on:/* 0! */   CELL(CELL(sp))=-1;sp+=4;break;
     case sw_0ST:/* 0! */  CELL(CELL(sp))=0;sp+=4;break;
     case sw_ST: /*  ! */  CELL(CELL(sp))=CELL(sp+4);sp+=8;break;
     case sw_addST: /*  +! */  CELL(CELL(sp))+=CELL(sp+4);sp+=8;break;
     case sw_cST:/* C! */  BYTE(CELL(sp))=CELL(sp+4);sp+=8;break;
     case sw_wST:/* W! */  WCELL(CELL(sp))=CELL(sp+4);sp+=8;break;
     case sw_2ST:/* 2! */    DCELL(CELL(sp))=DCELL(sp+4);sp+=12;break;
     case sw_2load:/* 2@ */ sp-=4; DCELL(sp)=DCELL(CELL(sp+4));break;
     case sw_2swap:/* 2swap */    dtmp=DCELL(sp);
			 DCELL(sp)=DCELL(sp+8);
			 DCELL(sp+8)=dtmp;break;
     case sw_cSTa:/* c!a  */ t=CELL(sp);BYTE(t)=CELL(sp+4);sp+=4;
                        CELL(sp)=t;break;

     case sw_qdup:/* dup  */  if(CELL(sp)==0)break;
     case sw_dup:/* dup  */ sp-=4;CELL(sp)=CELL(sp+4);break;
     case sw_2dup:/* 2dup */ sp-=4;CELL(sp)=CELL(sp+8);
     case sw_over:/* over */ sp-=4;CELL(sp)=CELL(sp+8);break;
     case sw_rload:/* r@   */ sp-=4;CELL(sp)=CELL(rp);break;
     case sw_rGT:/* r>   */   sp-=4;CELL(sp)=CELL(rp);rp+=4;break;
     case sw_GTr:/* >r   */   rp-=4;CELL(rp)=CELL(sp);sp+=4;break;
     case sw_ip:/*ip*/
    	 sp-=4;CELL(sp)=ip;ip+=4;break;
     case sw_lit:/*lit*/
    	 sp-=4;CELL(sp)=CELL(ip);ip+=4;break;

     case sw_litcomp:/*lit*/   	 	WCOM(sw_lit<<1|1);
     case sw_comp:/* , */    	 	COM(CELL(sp));    	 	sp+=4;break;
     case sw_Wcomp:/* W, */   	 	WCOM(CELL(sp));    	 	sp+=4;break;
     case sw_Ccomp:/* C, */    	 	CCOM(CELL(sp));    	 	sp+=4;break;
     case sw_allot:/* allot */   HERE+=CELL(sp); sp+=4;break;
     case sw_execute:/* execute */
    	 rp-=4;CELL(rp)=ip; ip=CELL(sp); sp+=4;break;
     case sw_compile:/* compile */
    	 Compile(CELL(sp));
    	 sp+=4;break;
     case sw_immediate:/* immediate */
    	 IMMED();break;
     case sw_femd:/*return*/  F_END();break;
     case sw_FromName:/*return*/ CELL(sp)=FromName(CELL(sp));break;


     case sw_spload: /* sp@ */ sp-=4;CELL(sp)=sp+4;break;
     case sw_spST: /* sp! */ sp=CELL(sp);break;
     case sw_rpload: /* rp@ */ sp-=4;CELL(sp)=rp;break;
     case sw_rpST: /* rp! */ rp=CELL(sp);sp+=4;break;

//     case sw_setalarm: /* setalarm */ setalarm(CELL(sp));sp+=4;break;

     case sw_to_catch:
   	 	 	    rp-=12;
   	 	 	    CELL(rp+8)=CELL(HANDLER);
   	 	 	    CELL(rp+4)=sp;
   	 	 	    CELL(rp)=ip;
   	 	 	    CELL(HANDLER)=rp;
   	 	 	 	ip=CELL(sp);
     	 	 	sp+=4;
     	 	 	 break;
     case sw_from_catch:
 	 	  	  	CELL(HANDLER)=CELL(rp+8-4);
 	 	  	  	rp+=12-4;
 	 	  	  	sp-=4;CELL(sp)=0;
 	 	  	  	break;


     case sw_throw:
    	 	 	 	if( !CELL(sp)) { sp+=4; break; }
    	 	 	 	if( !CELL(HANDLER)) {exit(CELL(sp));break;}
    	 	 	 	{ UNS32 savesp=CELL(sp);
    	 	 	 	  rp=CELL(HANDLER);
    	 	 	 	  CELL(HANDLER)=CELL(rp+8);
    	 	 	 	  sp=CELL(rp+4);
    	 	 	 	  CELL(sp)=savesp;
    	 	 	 	  ip=CELL(rp)+2; // over from_catch
    	 	 	 	  rp+=12;
    	 	 	 	 break;
    	 	 	 	}

     case sw_bye: /*exit*/  if(netid&0x80000000) closesocket(netid^0x80000000) ; return;
     case sw_branch: /*branch*/
    	 ip += (WCELL(ip)^0x8000) - 0x8000;break;
     case sw_0branch: /*0branch*/
    	 if(CELL(sp)) ip+=2; else ip+= (WCELL(ip)^0x8000) - 0x8000; sp+=4; break;

    	 case sw_BEGIN:  BEGIN();break;
    	 case sw_AGAIN:  AGAIN();break;
    	 case sw_UNTIL:  UNTIL();break;
    	 case sw_AHEAD:  AHEAD();break;
    	 case sw_IF:     IF();break;
    	 case sw_WHILE:  WHILE();break;
    	 case sw_THEN:   THEN();break;
    	 case sw_ELSE:   ELSE();break;
    	 case sw_REPEAT: REPEAT();break;


     case sw_key: /*read char */ sp-=4;
     CELL(sp)= getch();

     break;
     case sw_emit: /*print char*/
         twrite(fileids[CELL(OUTPUT)],&BYTE(sp),1);
    	 sp+=4;break;
     case sw_rdch: /*RH*/
        sp-=8; CELL(sp+4)=0;
        CELL(sp)=recv(fileids[3],&BYTE(sp+4),1,0);
    	 break;
     case sw_keyq: /*check key */ sp-=4;CELL(sp)=-kbhit();break;
     case sw_type: /*type*/
         if( CELL(sp+4)+CELL(sp) < MEMSIZE )
         twrite(fileids[CELL(OUTPUT)],&BYTE(CELL(sp+4)),CELL(sp));
        sp+=8;break;
     case sw_hdot: /*print hex*/
     {   char buffer[44];
         sprintf (buffer, "%X ", CELL(sp));
         twrite(fileids[CELL(OUTPUT)],buffer,strlen(buffer));
     }
     sp+=4;break;
     case sw_decdot: /*print decimal*/
     {   char buffer[44];
         sprintf (buffer, "%d ", CELL(sp));
         twrite(fileids[CELL(OUTPUT)],buffer,strlen(buffer));
     }
     sp+=4;break;
     case sw_dir:
     {
    	  DIR *dir;
    	  struct dirent *entry;
    	  char dirname[256];
    	  char flname[256];
    	  len=CELL(sp);addr=CELL(sp+4);
    	   if(!make_name(mem+addr,len)) {
     		  twrite(fileids[CELL(OUTPUT)],".. is unacceptable",18);
    		   break;}
    	  strcpy(dirname,getenv(HPATH));
    	  strcat(dirname,fpath);
    	  strcat(dirname,filename);
    	  strcat(dirname,"/");
	     	 { char buffer[144];
    	  sprintf (buffer,"%s\n",dirname);
    		 write(1,buffer,strlen(buffer));
    		 twrite(fileids[CELL(OUTPUT)],buffer,strlen(buffer));
    	 }
    	  if ((dir = opendir(dirname)) == NULL)
    		  twrite(fileids[CELL(OUTPUT)],"opendir() error",15);
    	  else {
    	    while ((entry = readdir(dir)) != NULL)
      	     {
      	     	 struct stat stbuf;
//      	     	 if(entry->d_name[0]!='.')
      	     	 {	 strcpy(flname,dirname);
      	     		 strcat(flname,entry->d_name);
      	     	 if(stat(flname, &stbuf)==0)
      	     	 { char buffer[44];
      	     		 sprintf (buffer,"%s\t%ld\n",entry->d_name,stbuf.st_size);
      	     		 twrite(fileids[CELL(OUTPUT)],buffer,strlen(buffer));
      	     	 }
      	     	 }
    	    }
    	    closedir(dir);
    	  }
     }  break;

     case sw_count: /*count*/ sp-=4;  CELL(sp)= BYTE(CELL(sp+4)); CELL(sp+4)++; break;
     case sw_str: /*(S")*/ sp-=8; t=BYTE(ip++); CELL(sp)=t; CELL(sp+4)=ip;
			ip+=t; if(ip&1)ip++; break;
     case sw_StrComp: /* S, */
			StrComp(CELL(sp+4),CELL(sp));
			sp+=8;break;
     case sw_NextWord:
    	 NextWord(); break;
     case sw_Parse:		Parse(); break;
     case sw_SQ: /* S" */ SQ();       break;
     case sw_PQ: /* ." */ SQ();	 WCOM(sw_type<<1|1); break;
     case sw_SourceST:		SourceST(); break;
     case sw_Source:		Source(); break;
     case sw_SearchWordList:	SearchWordList(); break;
     case sw_SFind:		SFind(); break;
     case sw_SHeader:	if( CELL(sp+4)+CELL(sp) < MEMSIZE )
    	                  SHEADER( &BYTE(CELL(sp+4)), CELL(sp) );
     	 	 	 	 sp+=8; break;
     case sw_openfile: /*open-file*/  sp+=4;len=CELL(sp);addr=CELL(sp+4);
                            CLIP SWAP if(!make_name(mem+addr,len)) {ior=-202;
                            SWAP CELL(sp)=ior;break;} SWAP
                            { int i;  char flname[256];
				flname[0]=0;
				 if(netid)
				 {  strcpy(flname,getenv(HPATH));
			    	    strcat(flname,fpath);
				 }

			    	    strcat(flname,filename);
	
                              for(i=6;i<20;i++) {
                              if(!fileids[i])break;
                              }
                              if(i==20){ior=-201;CELL(sp)=ior;break;}
//				printf("flname=<%s>\n",flname);
                              fp=open(flname,CELL(sp-4),S_IRUSR | S_IWUSR );
                              if(fp<0)ior=200;
                              fileids[i]=fp;
                              CELL(sp+4)=i;
                            }CELL(sp)=ior;break;
     case sw_closefile: /*close-file*/ FILEID(CELL(sp)) QFERR close(fp);
                            fileids[CELL(sp)]=0;
     			 CELL(sp)=ior;break;;
     case sw_writeline: /*write-line*/ sp+=8;FILEID(CELL(sp-8)) len=CELL(sp-4);
                            addr=CELL(sp);CLIP
				twrite(fp,mem+addr,len);
	    			twrite(fp,"\n",1);    ior=0;
                            CELL(sp)=ior;break;
     case sw_readline: /*read-line*/
    	 FILEID(CELL(sp))
    		 len=CELL(sp+4)+2;
             addr=CELL(sp+8);CLIP SWAP
				res=0;
//             printf("rl=%x %x\n",CELL(sp),fp );
		while(len--)
		{  int rrr=0;
			do{ rrr=tread(fp,&BYTE(addr+res),1);
				if(rrr==0)
				if((fp&0x80000000)==0)
				if(eof(fp)) rrr=-2;
				if(fp&0x80000000)
				{   if(rrr==-1)	return;
				    if(rrr==0)	return;
				}

				}while(rrr==-1);
				  if(rrr!=1) break;
				if(BYTE(addr+res)=='\n') { res++; break; }
				if(BYTE(addr+res)==8)
				{ if(res>1) res--;
					res--;
			}	res++;
		}   	ior=0;
                  if(res)CELL(sp+4)=0xffffffff;
                  else  CELL(sp+4)=0;
            		while((BYTE(addr+res-1)<' ')&(res>0)) res--;

                            CELL(sp+8)=res;
                            SWAP CELL(sp)=ior;break;
     case sw_writefile: /*write-file*/sp+=8;FILEID(CELL(sp-8)) len=CELL(sp-4);
                            addr=CELL(sp);CLIP
				twrite(fp,mem+addr,len);ior=0;
                            CELL(sp)=ior;break;
     case sw_readfile: /*read-file*/ sp+=4;FILEID(CELL(sp-4)) len=CELL(sp);
     	 	 	 	 addr=CELL(sp+4);CLIP
//     	 	 	 	 printf("rf=%x %x\n",CELL(sp-4),fp );
     	 	 	 	 res=tread(fp,mem+addr,len);ior=0;
     	 	 	 	 CELL(sp+4)=res;
			CELL(sp)=ior;break;
     case sw_deletefile: /*delete-file*/ sp+=4;len=CELL(sp-4);addr=CELL(sp);
                            CLIP SWAP if(!make_name(mem+addr,len)) {ior=-202;
                            SWAP CELL(sp)=ior;break;} SWAP remove(filename); CELL(sp)=ior;break;
     case sw_repositionfile: /*reposition-file*/ sp+=8;FILEID(CELL(sp-8)) QFERR
                             lseek(fp,CELL(sp),SEEK_SET);CELL(sp)=ior;break;

     case sw_fileposition: /*file-position*/ sp-=8;FILEID(CELL(sp+8)) QFERR
                             CELL(sp+8)=lseek(fp, (off_t) 0, SEEK_CUR);CELL(sp+4)=0;
                             CELL(sp)=ior;break;
     case sw_system: /*system */   sp+=4;len=CELL(sp-4);addr=CELL(sp);
                            CLIP SWAP if(!make_name(mem+addr,len)) {ior=-202;
                             SWAP CELL(sp)=ior;break;} SWAP ior=system(filename); CELL(sp)=ior;break;
     case sw_filesize: /* file-size*/ sp-=8;FILEID(CELL(sp+8))
     { int fpt;
         fpt=lseek(fp, (off_t) 0, SEEK_CUR);
         CELL(sp+8)=lseek(fp, (off_t) 0, SEEK_END);
         CELL(sp+4)=0;
            lseek(fp, (off_t) fpt, SEEK_SET);
      }break;
     case sw_restore: /* restore-system*/
      {  FILE *infile;
      	  	  sp+=4;len=CELL(sp-4);addr=CELL(sp);
              CLIP SWAP if(!make_name(mem+addr,len)) {ior=-202;
               SWAP CELL(sp)=ior;break;}
//      	  if((infile=fopen("forth.img","rb")))  fread(&CELL(0),1,MEMSIZE,infile);
      	  if((infile=fopen(filename,"rb")))  fread(&CELL(0),1,MEMSIZE,infile);
      }  zzz=1;
       printf("sw_restore\n");
       ip=FSTART; ireg=0;
       rp=MEMSIZE;
       break;

     case sw_telid:
    	 sp-=4;CELL(sp)=TELID; 	 break;

     case sw_ipadr:
		strncpy(inet_addr_buf,&BYTE(CELL(sp+4)),CELL(sp)&0xff);
		sp+=8;
		break;
     case sw_connect:
         { UNS32 tid=TELID;
          for(TELID=6;TELID<20;TELID++) {
           if(!fileids[TELID])break;
          }
          if(TELID==20){CELL(sp)=-1; TELID=tid; break;}
          if(CCNN!=-1) close(CCNN);
          CCNN=c_lconnect(CELL(sp))|0x80000000;
          fileids[TELID]=CCNN;
          CELL(sp)=TELID;
         }
    	 break;
     case sw_tellen:
		{  int nchar;
			sp-=4;
		if( ioctlsocket( CCNN^0x80000000, FIONREAD, &nchar ) == -1)
			{ CELL(sp)=0; break;
			} CELL(sp)=nchar;
		}
    	 break;
     case sw_telp:
     {
    	 int telpMAX = CELL(sp);
    	 int telpBUF = CELL(sp+4);
    	 int telpCount = 0;
			for(;;)
			{ char cc; int rrr; int nchar;
			if( ioctlsocket( CCNN^0x80000000, FIONREAD, &nchar ) == -1) break;
                        if(nchar==0)  { CELL(sp)=telpCount; break; }
			  rrr=tread(CCNN,&cc,1);
			  if(rrr==0) { CELL(sp)=telpCount;  break;}
			  if(rrr==-1) { CELL(sp)=telpCount; break; }
			  if(calltelp)
			  {
//				  twrite(fileids[CELL(OUTPUT)],"~",1);
				  if( cc=='\n' || telpCount==telpMAX )
				  { CELL(sp)=telpCount; calltelp=0; break;
				  }
//				  twrite(fileids[CELL(OUTPUT)],&cc,1);
				  BYTE(telpBUF+telpCount)=cc;
				  telpCount++;
			  }
			  else
			  {	  if(tplast=='\n' && cc=='~')
				  {	calltelp=1;
				  }
			  	  else
			  	  {	  twrite(fileids[CELL(OUTPUT)],&cc,1);
			  	  	  tplast=cc;
			  	  }
			  }
			}
     }
	    	  break;

     case sw_ms: /* ms*/ usleep(CELL(sp)*1000);sp+=4;break;

     case sw_hi:
         twrite(fileids[CELL(OUTPUT)],"Hello!!!\n",9);
         break;
break;
    }
   if(ireg&0xC000)/*return*/ {
            ip=CELL(rp);
            rp+=4;
           } 

  }
 }
}

make_image()
{ 	MEMDEF
	UNS32 shere;
  HERE=0;
 HERE=vocs_beg;
 CELL(HANDLER)=0;
 CELL(ATIB)=TTIB;
 CELL(CONTEXT)=VFORTH;
 CELL(CONTEXT+4)=0;
 CELL(CURRENT)=VFORTH;
 CELL(INPUT)=0;
 CELL(OUTPUT)=1;
 current=0;

write(1,"make_image\n",11);
PRIM("NOOP",sw_nop);
PRIM("CHARS",sw_nop);
PRIM("SWAP",sw_swap);
PRIM("ROT",sw_rot);
PRIM("0=",sw_0equ);
PRIM("0<>",sw_0Nequ);
PRIM("0<",sw_0LT);
PRIM("0>",sw_0GT);
PRIM("=",sw_equ);
PRIM("<>",sw_nequ);
PRIM("NEGATE",sw_negate);
PRIM("INVERT",sw_invert);
PRIM("UM*",sw_ummul);
PRIM("C@",sw_Cload);
PRIM("W@",sw_Wload);
PRIM("@",sw_load);
PRIM("ON",sw_on);
PRIM("0!",sw_0ST);
PRIM("1+",sw_1add);
PRIM("2+",sw_2add);
PRIM("1-",sw_1sub);
PRIM("2-",sw_2sub);
PRIM("2*",sw_2mul);
PRIM("2/",sw_2div);
PRIM("+",sw_add);
PRIM("-",sw_sub);
PRIM("*",sw_mul);
PRIM("/",sw_div);
PRIM("MOD",sw_mod);
PRIM("CELL/",sw_celldiv);
PRIM("CELLS",sw_cells);
PRIM("AND",sw_and);
PRIM("OR",sw_or);
PRIM("XOR",sw_xor);
PRIM("U<",sw_uLT);
PRIM("<",sw_LT);
PRIM("U>",sw_uGT);
PRIM(">",sw_GT);
PRIM("MIN",sw_min);
PRIM("MAX",sw_max);
PRIM("LSHIFT",sw_lshift);
PRIM("RSHIFT",sw_rshift);
PRIM("UM/MOD",sw_umSmod);
PRIM("+CY",sw_addcy);
PRIM("D+",sw_Dadd);
PRIM("SCAN1",sw_scan1);
PRIM("DROP",sw_drop);
PRIM("2DROP",sw_2drop);
PRIM("LIT,",sw_litcomp);
PRIM(",",sw_comp);
PRIM("W,",sw_Wcomp);
PRIM("C,",sw_Ccomp);
PRIM("ALLOT",sw_allot);
PRIM("EXECUTE",sw_execute);
PRIM("COMPILE,",sw_compile);
PRIM("IMMEDIATE",sw_immediate);
PRIM("NAME>",sw_FromName);
PRIM(">BODY",sw_2add);
PRIM("SP@",sw_spload);
PRIM("SP!",sw_spST);
PRIM("RP@",sw_rpload);
PRIM("RP!",sw_rpST);

PRIM("THROW",sw_throw);
PRIM("W,",sw_Wcomp);
PRIM("!",sw_ST);
PRIM("+!",sw_addST);
PRIM("C!",sw_cST);
PRIM("W!",sw_wST);
PRIM("2!",sw_2ST);
PRIM("2@",sw_2load);
PRIM("2SWAP",sw_2swap);
PRIM("C!A",sw_cSTa);
PRIM("?DUP",sw_qdup);
PRIM("DUP",sw_dup);
PRIM("2DUP",sw_2dup);
PRIM("OVER",sw_over);
PRIM("R@",sw_rload);
PRIM("R>",sw_rGT);
PRIM(">R",sw_GTr);
PRIM("KEY",sw_key);
PRIM("EMIT",sw_emit);
PRIM("RC",sw_rdch);
PRIM("KEY?",sw_keyq);
PRIM("TYPE",sw_type);
PRIM("HI",sw_hi);
PRIM("BYE",sw_bye);
PRIM("H.",sw_hdot);
PRIM("DEC.",sw_decdot);
PRIM("$DIR",sw_dir);
PRIM("COUNT",sw_count);
PRIM("PARSE-NAME",sw_NextWord);
PRIM("PARSE",sw_Parse);
PRIM("S\"", sw_SQ); IMMED();
PRIM(".\"", sw_PQ); IMMED();
PRIM("SOURCE",sw_Source);
PRIM("SOURCE!",sw_SourceST);
PRIM("SEARCH-WORDLIST",sw_SearchWordList);
PRIM("SFIND",sw_SFind);
PRIM("SHEADER",sw_SHeader);

PRIM("BEGIN",sw_BEGIN); IMMED();
PRIM("AGAIN",sw_AGAIN); IMMED();
PRIM("UNTIL",sw_UNTIL); IMMED();
PRIM("AHEAD",sw_AHEAD); IMMED();
PRIM("IF",sw_IF); IMMED();
PRIM("WHILE",sw_WHILE); IMMED();
PRIM("THEN",sw_THEN); IMMED();
PRIM("ELSE",sw_ELSE); IMMED();
PRIM("REPEAT",sw_REPEAT); IMMED();


PRIM("OPEN-FILE",sw_openfile);
PRIM("CLOSE-FILE",sw_closefile);

PRIM("WRITE-LINE",sw_writeline);
PRIM("READ-LINE",sw_readline);

PRIM("WRITE-FILE",sw_writefile);
PRIM("READ-FILE",sw_readfile);

PRIM("DELETE-FILE",sw_deletefile);
PRIM("REPOSITION-FILE",sw_repositionfile);
PRIM("FILE-POSITION",sw_fileposition);
PRIM("SYSTEM",sw_system);
PRIM("FILE-SIZE",sw_filesize);
PRIM("RESTORE-SYSTEM",sw_restore);
//PRIM("RS",sw_restore);
PRIM("TELID",sw_telid);
PRIM("IPADR!",sw_ipadr);
PRIM("CONNECT",sw_connect);
PRIM("TELN",sw_tellen);
PRIM("TEL.",sw_telp);
PRIM("MS",sw_ms);

CONSTANT("CELL",4);
CONSTANT(">IN",TO_IN);
CONSTANT("STATE",STATE);
CONSTANT("[EXIT]",0x8001);
CONSTANT("C/L",TIB_SAZE);
CONSTANT("ATIB",ATIB);
CONSTANT("TTIB",TTIB);
CONSTANT("#TIB",NTIB);
CONSTANT("DP",DPOINT);
CONSTANT("CURRENT",CURRENT);
CONSTANT("CONTEXT",CONTEXT);
CONSTANT("LAST",LAST);
CONSTANT("FORTH-WORDLIST",VFORTH);
CONSTANT("CREATE-CODE",sw_ip<<1|0x4001);
CONSTANT("CONSTANT-CODE",sw_lit<<1|0x4001);

CONSTANT("R/O",O_RDONLY);
CONSTANT("W/O",O_WRONLY);
CONSTANT("R/W",O_RDWR);
CONSTANT("O_CREAT",O_CREAT);

HERE+=0x300;

CONSTANT("PAD",HERE-0x100);

VARIABLE("DPL");
VARIABLE("BASE");
VARIABLE("SP0");
CONSTANT("STDOUT",OUTPUT);

HEADER("DECIMAL"); LIT(10); CC("BASE");  CC("!"); F_END();
HEADER("HEX");     LIT(16); CC("BASE");  CC("!"); F_END();
HEADER("["); CC("STATE");  CC("0!"); F_END();  IMMED();
HEADER("]"); CC("STATE");  CC("ON"); F_END();
HEADER("TIB"); CC("ATIB");  CC("@"); F_END();


HEADER("CATCH");
WCOM(sw_to_catch<<1|1);
WCOM(sw_from_catch<<1|1);
F_END();

HEADER("TST"); LIT(0x33); CC("H.");
 F_END();

 HEADER("DNEGATE"); // ( d1 --- d2)
 // Negate the top double number on the stack.
 CC(">R"); CC("NEGATE"); CC("R>"); CC("NEGATE"); CC("OVER"); CC("0<>");  CC("+");
 F_END();


 HEADER("DIGIT?"); // ( c -- 0| c--- n -1)
 // Convert character c to its digit value n and return true if c is a
 // digit in the current base. Otherwise return false.
 LIT('0'); CC("-"); CC("DUP"); CC("0<"); IF(); CC("DROP"); LIT(0); WCOM(0x8001); THEN();
 CC("DUP"); LIT(9); CC(">"); CC("OVER"); LIT(17); CC("<"); CC("AND"); IF(); CC("DROP"); LIT(0); WCOM(0x8001); THEN();
 CC("DUP"); LIT(9); CC(">"); IF(); LIT(7); CC("-"); THEN();
 CC("DUP"); CC("BASE"); CC("@"); CC("<"); CC("0="); IF(); CC("DROP"); LIT(0); WCOM(0x8001); THEN();
 LIT(-1);
 F_END();

 HEADER(">NUMBER"); // ( ud1 c-addr1 u1 --- ud2 c-addr2 u2 )
 // Convert the string at c-addr with length u1 to binary, multiplying ud1
 // by the number in BASE and adding the digit value to it for each digit.
 // c-addr2 u2 is the remainder of the string starting at the first character
 // that is no digit.
   BEGIN();
   CC("DUP");
   WHILE();
    CC("1-"); CC(">R");
    CC("COUNT"); CC("DIGIT?"); CC("0=");
    IF();
    CC("R>"); CC("1+"); CC("SWAP"); CC("1-"); CC("SWAP");  WCOM(0x8001);
    THEN();
    CC("SWAP"); CC(">R");
    CC(">R");
    CC("SWAP"); CC("BASE"); CC("@"); CC("UM*"); CC("ROT"); CC("BASE"); CC("@");
    CC("*"); LIT(0); CC("SWAP"); CC("D+"); // Multiply ud by base.
    CC("R>"); LIT(0); CC("D+");           // Add new digit.
    CC("R>"); CC("R>");
   REPEAT();
   F_END();

 HEADER("CONVERT"); // ( ud1 c-addr1 --- ud2 c-addr2)
 // Convert the string starting at c-addr1 + 1 to binary. c-addr2 is the
 // address of the first non-digit. Digits are added into ud1 as in >NUMBER
 CC("1-"); LIT(-1); CC(">NUMBER"); CC("DROP");
 F_END();

 HEADER("NUMBER?"); // ( c-addr n ---- d f)
 // Convert the counted string at c-addr to a double binary number.
 // f is true if and only if the conversion was successful. DPL contains
 // -1 if there was no point in the number, else the position of the point
 // from the right. Special prefixes: # means decimal, $ means hex.
 LIT(-1); CC("DPL"); CC("!");
 CC("BASE"); CC("@"); CC(">R");
// CC("COUNT");
 CC("OVER"); CC("C@"); LIT(45); CC("="); CC("DUP"); CC(">R"); IF(); CC("1-"); CC("SWAP"); CC("1+"); CC("SWAP"); THEN(); // Get any - sign
 CC("OVER"); CC("C@"); LIT(36); CC("="); IF(); LIT(16); CC("BASE"); CC("!"); CC("1-"); CC("SWAP"); CC("1+"); CC("SWAP"); THEN();   // $ sign for hex.
 CC("OVER"); CC("C@"); LIT(35); CC("="); IF(); LIT(10); CC("BASE"); CC("!"); CC("1-"); CC("SWAP"); CC("1+"); CC("SWAP"); THEN();   // # sign for decimal
 CC("DUP");  CC("0>"); CC("0=");
 IF();  CC("R>"); CC("DROP"); CC("R>"); CC("BASE"); CC("!"); LIT(0); WCOM(0x8001);
 THEN();   // Length 0 or less?
 CC(">R"); CC(">R"); LIT(0); LIT(0); CC("R>"); CC("R>");
   BEGIN();
 CC(">NUMBER");
 CC("DUP"); IF(); CC("OVER"); CC("C@"); LIT(46); CC("="); IF(); CC("1-"); CC("DUP");
 CC("DPL"); CC("!"); CC("SWAP"); CC("1+"); CC("SWAP"); ELSE(); // handle point.
 CC("R>"); CC("DROP"); CC("R>"); CC("BASE"); CC("!"); LIT(0); WCOM(0x8001); THEN();  // Error if anything but point
        THEN();
   CC("DUP"); CC("0="); UNTIL(); CC("2DROP"); CC("R>"); IF(); CC("DNEGATE"); THEN();
   CC("R>"); CC("BASE"); CC("!"); LIT(-1);
   F_END();

HEADER("DEPTH");
  CC("SP0"); CC("@"); CC("SP@");  CC("-");
  CC("CELL/"); CC("1-");
  F_END();

HEADER("?STACK");
CC("DEPTH");  CC("0<");
IF(); LIT(-4); CC("THROW");
THEN();
F_END();

 HEADER("INTERPRET");
// TP("INTERPRET\n");
 BEGIN();
 CC("PARSE-NAME"); CC("DUP");
 WHILE();
 CC("SFIND"); CC("?DUP");
    IF();
    CC("STATE"); CC("@"); CC("=");
    	IF();  CC("COMPILE,");
    	ELSE();
//    	  TP("st=");  CC("DUP");  CC("H.");
    	CC("EXECUTE");
    	THEN();
    ELSE();

    CC("NUMBER?"); CC("0="); IF(); LIT(-13); CC("THROW"); THEN();

    CC("DPL"); CC("@"); CC("1+"); IF();
    CC("STATE"); CC("@"); IF(); CC("SWAP"); CC("LIT,"); CC("LIT,"); THEN();
       ELSE();
       CC("DROP"); CC("STATE"); CC("@"); IF(); CC("LIT,"); THEN();
       THEN();

    THEN();

 CC("?STACK");
 REPEAT();
 CC("2DROP");
 F_END();

 HEADER("EVAL");
 CC("SOURCE!");
 CC(">IN");   CC("0!");
 CC("INTERPRET");
 F_END();

 HEADER(":");
 CC("STATE"); CC("ON");
 CC("PARSE-NAME");
// TP("PARSE-NAME=");  CC("2DUP"); CC("TYPE");
 CC("SHEADER");
 F_END();

 HEADER("EXIT");  IMMED();
 CC("[EXIT]"); CC("W,");
 F_END();

 HEADER(";");  IMMED();
 WCOM(sw_femd<<1|1);
 CC("STATE"); CC("0!");
 F_END();

 HEADER("'");
 CC("PARSE-NAME"); CC("SFIND");  CC("0=");
 IF(); LIT(-321); CC("THROW");
 THEN();
 F_END();

 HEADER("[']"); IMMED();
 CC("'"); CC("LIT,");
 F_END();

 HEADER("(");   IMMED();
  LIT(')'); CC("PARSE"); CC("2DROP");
 F_END();

 HEADER(".(");   IMMED();
  LIT(')'); CC("PARSE"); CC("TYPE");
 F_END();

 HEADER("\\");   IMMED();
  LIT(0); CC("PARSE"); CC("2DROP");
 F_END();

 HEADER("CR");
  LIT('\n');  CC("EMIT");
 F_END();

 HEADER("SPACE");
  LIT(' ');  CC("EMIT");
 F_END();

 HEADER("MU/MOD");
 // ( ud u --- urem udquot )  Divide unsigned double number
 //ud by u and return a double quotient and G a single remainder.
 CC(">R"); LIT(0); CC("R@"); CC("UM/MOD"); CC("R>");
       CC("SWAP"); CC(">R"); CC("UM/MOD"); CC("R>");
F_END();

 HEADER("DABS"); // ( d --- ud)  ud is the absolute value of d.
		 CC("DUP"); CC("0<"); IF(); CC("DNEGATE"); THEN();
F_END();

 VARIABLE("HLD");

 HEADER("HOLD"); // ( c ---)  Insert character c into the  numerical  conver- sion buffer.
 LIT(-1);  CC("HLD"); CC("+!"); CC("HLD"); CC("@"); CC("C!");
 F_END();

 HEADER("#");  // ( ud1 --- ud2)  Extract the rightmost digit of ud1 and put
 // it into the numerical G conversion buffer.
 CC("BASE"); CC("@"); CC("MU/MOD"); CC("ROT"); CC("DUP");
  LIT(9); CC(">");
  IF(); LIT(7); CC("+");
  THEN(); LIT(48); CC("+"); CC("HOLD");
 F_END();

 HEADER("#S");  // ( ud --- 0 0 )  Convert ud by repeated use of # until ud is zero.
   BEGIN(); CC("#"); CC("2DUP"); CC("OR"); CC("0="); UNTIL();
 F_END();

 HEADER("SIGN"); // (  n  ---)  Insert a - sign in the numerical conversion buffer if n is negative.
		 CC("0<"); IF(); LIT(45); CC("HOLD"); THEN();
 F_END();

 HEADER("<#"); // ( --- )  Reset the numerical conversion buffer.
 CC("PAD"); CC("HLD"); CC("!");
 F_END();

 HEADER("#>"); // ( ud --- addr u )  Discard ud  and  give  the  address  and  length of the numerical conversion G buffer.
 CC("2DROP");  CC("HLD"); CC("@"); CC("PAD"); CC("OVER"); CC("-");
     F_END();

 HEADER("(D.)");
 CC("SWAP"); CC("OVER"); CC("DABS"); CC("<#"); CC("#S"); CC("ROT"); CC("SIGN");
 CC("#>");
     F_END();

HEADER("D."); // ( d --- )  Type the double number d to the terminal.
  CC("(D.)"); CC("TYPE"); CC("SPACE");
  F_END();

 HEADER("U."); // ( u ---)  Type the unsigned number u to the terminal.
     LIT(0); CC("D.");
     F_END();

 HEADER("S>D");
   CC("DUP");  CC("0<");
 F_END();

 HEADER("."); // ( n ---)  Type the signed number n to the terminal.
		 CC("S>D"); CC("D.");
 F_END();

 VARIABLE("SID");


 HEADER("SOURCE-ID"); CC("SID");  CC("@");
 F_END();

HEADER("\\EOF");
		 CC("SOURCE-ID"); CC("FILE-SIZE"); CC("DROP");
		 CC("SOURCE-ID"); CC("REPOSITION-FILE"); CC("DROP");
		 CC("\\");
F_END();


 HEADER("ACCEPT"); // TP("READ-LINE");
  LIT(0); CC("READ-LINE"); CC("2DROP");
 F_END();

 HEADER("REFILL");
 CC("TIB"); CC("C/L"); CC("SOURCE-ID");
// CC("DUP"); CC("H.");   TP("READ-LINE\n");
   CC("READ-LINE");  CC("DROP");
   IF();  CC("TIB"); CC("SWAP"); CC("SOURCE!");  LIT(-1);
   ELSE(); CC("DROP"); LIT(0);
   THEN();
   CC(">IN"); CC("0!");
   F_END();

   HEADER("QUIT");
//   TP("BEGIN QUIT\n");
   BEGIN();    CC("REFILL");
   WHILE();    CC("INTERPRET");
   REPEAT();
   F_END();
  shere=HERE;
  HERE+=1000;

 FSTART=HERE;
 TP("Forth\n");
 CC("STATE"); CC("0!");
 CC("SID"); CC("0!");
 CC("SP@"); CC("SP0"); CC("!");
CC("DECIMAL");
 //CC("SP@");  CC("SP0"); CC("!");

STR(" .( DEPTH=) DEPTH . CR");  CC("EVAL");

 STR(": INCLUDE-FILE");  CC("EVAL"); // ( fid --- )
 // Read lines from the file identified by fid and interpret them.
 // INCLUDE and EVALUATE nest in arbitrary order.
 STR(" SID @ >R >IN @ >R ATIB @ >R #TIB @ >R RP@ C/L - RP! ");  CC("EVAL");
 STR(" SID ! RP@ ATIB ! ");  CC("EVAL");

 STR(" ['] QUIT CATCH ");  CC("EVAL");
 STR(" DUP IF CR SOURCE TYPE CR THEN");  CC("EVAL");

 STR(" RP@ C/L + RP! R> #TIB ! R> ATIB ! R> >IN ! R> SID ! THROW ;");  CC("EVAL");

 STR(": INCLUDED");  CC("EVAL"); // ( c-addr u ---- )
 // Open the file with name c-addr u and interpret all lines contained in it.
 STR("R/O OPEN-FILE IF -38 THROW THEN");  CC("EVAL");
 STR("DUP >R ['] INCLUDE-FILE CATCH");  CC("EVAL");
 STR("R> CLOSE-FILE DROP THROW ;");  CC("EVAL");

 STR(": ?ERR. ?DUP IF CR .\" ERR=\" DEC. CR THEN ;");  CC("EVAL");
 STR(": MAIN BEGIN STATE 0! ['] QUIT CATCH ?ERR. SP0 @ SP! AGAIN ;");  CC("EVAL");
 STR(": AUTOEXEC S\" autoexec.4\" ['] INCLUDED CATCH ?ERR. SP0 @ SP! STATE 0! ;"); CC("EVAL");
 STR("AUTOEXEC MAIN"); CC("EVAL");
// S" autoexec.4" INCLUDED
 CC("BYE");
 HERE=shere;

}

/* Free the thread-specific buffer */
 static void buffer_destroy(void * buf)
 {
   free(buf);
 }

 /* Allocate the key */
  static void buffer_key_alloc()
  {
    pthread_key_create(&buffer_key, buffer_destroy);
  }


volatile UNS8 * mainmem=NULL;

void *vfm_do(void *pnetid)
{	UNS8 * mem,mem1;
	FILE *infile;
//   mem1=malloc(1000);   free(mem1);
   pthread_setspecific(buffer_key, mem=malloc(MEMSIZE+3+4));
 sp=MEMSIZE-2*1024;
 if(0) //(infile=fopen("forth.img","rb")))
 {
	 fread(mem,1,MEMSIZE,infile);
 }else

 if(mainmem)
 {
	 memcpy(mem,mainmem,MEMSIZE);
}
 else{
	 make_image();
 mainmem=mem;
// printf("mainmem=mem\n");
 }

 virtual_machine( *(UNS32 *) pnetid );

 printf("virtual_machine exit\n");

}


pthread_attr_t attr;
main(int count, char *strings[])
{ char ii;
pthread_t pthread;
UNS32 netid;
UNS8 * mem1;
char dirname[256];
if ( count != 2 )
{
    printf("Usage: %s <portnum>\n", strings[0]);
    exit(0);
}

    	  strcpy(dirname,getenv(HPATH));
    	  strcat(dirname,"/FPPublic");
#ifdef WIN32
	mkdir(dirname);
#else
	mkdir(dirname,0777);
#endif
    	  strcpy(dirname,getenv(HPATH));
    	  strcat(dirname,"/.FPauk");
#ifdef WIN32
	mkdir(dirname);
#else
	mkdir(dirname,0777);

//	signal(SIGINT, SIGINTTypt );
//	signal(SIGALRM, SIGALRMTypt );

	signal(SIGPIPE, SIGSIGPIPETypt);
//	signal(SIGTERM, SIGSIGTERMTypt);
//	signal(SIGSEGV, SIGSIGSEGVTypt);

#endif
//portnum = strings[1];

 int sss=c_serv(atoi(strings[1]));

// mem1=malloc(1000); free(mem1);

   pthread_attr_init(&attr);
   pthread_once(&buffer_key_once, buffer_key_alloc);

 netid=0;
  for(;;)
 {
  pthread_create(&pthread, &attr, vfm_do, (void *)&netid);
 netid=c_accept(sss)|0x80000000;
//  for(;;);
 }
 return 0;
}
